declare type Flag = 'd' | 'g' | 'i' | 'm' | 's' | 'u' | 'y';
/** Generate indices for substring matches */
declare const withIndices = "d";
/** Case-insensitive search */
declare const caseInsensitive = "i";
/** Global search */
declare const global = "g";
/** Multi-line search */
declare const multiline = "m";
/** Allows `.` to match newline characters */
declare const dotAll = "s";
/** Treat a pattern as a sequence of unicode code points */
declare const unicode = "u";
/** Perform a "sticky" search that matches starting at the current position in the target string */
declare const sticky = "y";

declare type InputSource<S extends string = never, T extends string = never> = S | Input<S, T>;
declare type GetGroup<T extends InputSource<string>> = T extends Input<string, infer Group> ? Group : never;
declare type MapToValues<T extends InputSource<any, any>[]> = T extends [infer First, ...infer Rest] ? First extends InputSource<string> ? [GetValue<First>, ...MapToValues<Rest>] : [] : [];
declare type MapToGroups<T extends InputSource<any, string>[]> = T extends [
    infer First,
    ...infer Rest
] ? First extends Input<any, infer K> ? K | MapToGroups<Rest> : MapToGroups<Rest> : never;

declare type ExactEscapeChar = '.' | '*' | '+' | '?' | '^' | '$' | '{' | '}' | '(' | ')' | '|' | '[' | ']' | '/';
declare type Escape<T extends string, EscapeChar extends string> = T extends `${infer Start}${EscapeChar}${string}` ? Start extends `${string}${EscapeChar}${string}` ? never : T extends `${Start}${infer Char}${string}` ? Char extends EscapeChar ? T extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Escape<Rest, EscapeChar>}` : never : never : never : T;
declare type StripEscapes<T extends string> = T extends `${infer A}\\${infer B}` ? `${A}${B}` : T;
declare type GetValue<T extends InputSource<string>> = T extends string ? Escape<T, ExactEscapeChar> : T extends Input<infer R> ? R : never;

declare type Wrap<T extends string, Yes, No> = T extends `(${string})` ? Yes : StripEscapes<T> extends `${infer A}${infer B}` ? A extends '' ? Yes : B extends '' ? Yes : No : never;

interface Input<V extends string, G extends string = never> {
    and: {
        /** this adds a new pattern to the current input */
        <I extends InputSource<string, any>>(input: I): Input<`${V}${GetValue<I>}`, G | (I extends Input<any, infer NewGroups> ? NewGroups : never)>;
        /** this adds a new pattern to the current input, with the pattern reference to a named group. */
        referenceTo: <N extends G>(groupName: N) => Input<`${V}\\k<${N}>`, G>;
    };
    /** this provides an alternative to the current input */
    or: <I extends InputSource<string, any>>(input: I) => Input<`(${V}|${GetValue<I>})`, G | (I extends Input<any, infer NewGroups> ? NewGroups : never)>;
    /** this is a positive lookbehind. Make sure to check [browser support](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/RegExp#browser_compatibility) as not all browsers support lookbehinds (notably Safari) */
    after: <I extends InputSource<string>>(input: I) => Input<`(?<=${GetValue<I>})${V}`, G>;
    /** this is a positive lookahead */
    before: <I extends InputSource<string>>(input: I) => Input<`${V}(?=${GetValue<I>})`, G>;
    /** these is a negative lookbehind. Make sure to check [browser support](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/RegExp#browser_compatibility) as not all browsers support lookbehinds (notably Safari) */
    notAfter: <I extends InputSource<string>>(input: I) => Input<`(?<!${GetValue<I>})${V}`, G>;
    /** this is a negative lookahead */
    notBefore: <I extends InputSource<string>>(input: I) => Input<`${V}(?!${GetValue<I>})`, G>;
    times: {
        /** repeat the previous pattern an exact number of times */
        <N extends number>(number: N): Wrap<V, Input<`${V}{${N}}`, G>, Input<`(${V}){${N}}`, G>>;
        /** specify that the expression can repeat any number of times, _including none_ */
        any: () => Wrap<V, Input<`${V}*`, G>, Input<`(${V})*`, G>>;
        /** specify that the expression must occur at least x times */
        atLeast: <N extends number>(number: N) => Wrap<V, Input<`${V}{${N},}`, G>, Input<`(${V}){${N},}`, G>>;
        /** specify a range of times to repeat the previous pattern */
        between: <Min extends number, Max extends number>(min: Min, max: Max) => Wrap<V, Input<`${V}{${Min},${Max}}`, G>, Input<`(${V}){${Min},${Max}}`, G>>;
    };
    /** this defines the entire input so far as a named capture group. You will get type safety when using the resulting RegExp with `String.match()` */
    as: <K extends string>(key: K) => Input<`(?<${K}>${V})`, G | K>;
    /** this allows you to match beginning/ends of lines with `at.lineStart()` and `at.lineEnd()` */
    at: {
        lineStart: () => Input<`^${V}`, G>;
        lineEnd: () => Input<`${V}$`, G>;
    };
    /** this allows you to mark the input so far as optional */
    optionally: () => Wrap<V, Input<`${V}?`, G>, Input<`(${V})?`, G>>;
    toString: () => string;
}

declare type Join<T extends string[], Prefix extends string = '', Joiner extends string = '|'> = T extends [infer F, ...infer R] ? F extends string ? `${Prefix}${F}${R extends string[] ? Join<R, Joiner, Joiner> : ''}` : '' : '';

/** This matches any character in the string provided */
declare const charIn: <T extends string>(chars: T) => Input<`[${T extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : T extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? T extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${any}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : T}]`, never>;
/** This matches any character that is not in the string provided */
declare const charNotIn: <T extends string>(chars: T) => Input<`[^${T extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : T extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? T extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${Rest extends `${infer Start}^${string}` | `${infer Start}]${string}` | `${infer Start}\\${string}` | `${infer Start}-${string}` ? Start extends `${string}^${string}` | `${string}]${string}` | `${string}\\${string}` | `${string}-${string}` ? never : Rest extends `${Start}${infer Char}${string}` ? Char extends "^" | "]" | "\\" | "-" ? Rest extends `${Start}${Char}${infer Rest}` ? `${Start}\\${Char}${any}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : Rest}` : never : never : never : T}]`, never>;
/** This takes an array of inputs and matches any of them. */
declare const anyOf: <New extends InputSource<V, T>[], V extends string, T extends string>(...args: New) => Input<`(${Join<MapToValues<New>, "", "|">})`, MapToGroups<New>>;
declare const char: Input<".", never>;
declare const word: Input<"\\b\\w+\\b", never>;
declare const wordChar: Input<"\\w", never>;
declare const wordBoundary: Input<"\\b", never>;
declare const digit: Input<"\\d", never>;
declare const whitespace: Input<"\\s", never>;
declare const letter: Input<"[a-zA-Z]", never>;
declare const tab: Input<"\\t", never>;
declare const linefeed: Input<"\\n", never>;
declare const carriageReturn: Input<"\\r", never>;
declare const not: {
    wordChar: Input<"\\W", never>;
    wordBoundary: Input<"\\B", never>;
    digit: Input<"\\D", never>;
    whitespace: Input<"\\S", never>;
    letter: Input<"[^a-zA-Z]", never>;
    tab: Input<"[^\\t]", never>;
    linefeed: Input<"[^\\n]", never>;
    carriageReturn: Input<"[^\\r]", never>;
};
/** Equivalent to `?` - this marks the input as optional */
declare const maybe: <New extends InputSource<string, never>>(str: New) => Wrap<GetValue<New>, Input<`${GetValue<New>}?`, GetGroup<New>>, Input<`(${GetValue<New>})?`, GetGroup<New>>>;
/** This escapes a string input to match it exactly */
declare const exactly: <New extends InputSource<string, never>>(input: New) => Input<GetValue<New>, GetGroup<New>>;
declare const oneOrMore: <New extends InputSource<string, never>>(str: New) => Wrap<GetValue<New>, Input<`${GetValue<New>}+`, GetGroup<New>>, Input<`(${GetValue<New>})+`, GetGroup<New>>>;

declare const NamedGroupsS: unique symbol;
declare const ValueS: unique symbol;
declare const FlagsS: unique symbol;
declare type MagicRegExp<Value extends string, NamedGroups extends string | never = never, Flags extends string | never = never> = RegExp & {
    [NamedGroupsS]: NamedGroups;
    [ValueS]: Value;
    [FlagsS]: Flags;
};
declare type ExtractGroups<T extends MagicRegExp<string, string, string>> = T extends MagicRegExp<string, infer V, string> ? V : never;
declare type MagicRegExpMatchArray<T extends MagicRegExp<string, string, string>> = Omit<RegExpMatchArray, 'groups'> & {
    groups: Record<ExtractGroups<T>, string | undefined>;
};

declare const createRegExp: <Value extends string, NamedGroups extends string = never, Flags extends Flag[] = never[]>(raw: Value | Input<Value, NamedGroups>, flags?: string | [...Flags] | Set<Flag> | undefined) => MagicRegExp<`/${Value}/${Join<Flags, "", "">}`, NamedGroups, Flags[number]>;

declare global {
    interface String {
        match<R extends MagicRegExp<string, string, Exclude<Flag, 'g'>>>(regexp: R): MagicRegExpMatchArray<R> | null;
        match<R extends MagicRegExp<string, string, 'g'>>(regexp: R): string[] | null;
        /** @deprecated String.matchAll requires global flag to be set. */
        matchAll<R extends MagicRegExp<string, string, never>>(regexp: R): never;
        /** @deprecated String.matchAll requires global flag to be set. */
        matchAll<R extends MagicRegExp<string, string, Exclude<Flag, 'g'>>>(regexp: R): never;
        matchAll<R extends MagicRegExp<string, string, string>>(regexp: R): IterableIterator<MagicRegExpMatchArray<R>>;
        /** @deprecated String.replaceAll requires global flag to be set. */
        replaceAll<R extends MagicRegExp<string, string, never>>(searchValue: R, replaceValue: string | ((substring: string, ...args: any[]) => string)): never;
        /** @deprecated String.replaceAll requires global flag to be set. */
        replaceAll<R extends MagicRegExp<string, string, Exclude<Flag, 'g'>>>(searchValue: R, replaceValue: string | ((substring: string, ...args: any[]) => string)): never;
    }
}

export { Flag, Input, MagicRegExp, MagicRegExpMatchArray, anyOf, carriageReturn, caseInsensitive, char, charIn, charNotIn, createRegExp, digit, dotAll, exactly, global, letter, linefeed, maybe, multiline, not, oneOrMore, sticky, tab, unicode, whitespace, withIndices, word, wordBoundary, wordChar };
